//*******************************************************
//
//	KTR-10
//
//	Si5351A 制御
//
//  参考：Si5351Aの使い方と動作 / Computer & RF Technology
//    https://ttrf.tk/posts/2016-09-11-si5351a-configuration-how-to-and-signal-quality/
//
//	2023.6.9	KAWAKAMI Yukio
//*******************************************************
#include "Si5351.h"

#define SI5351_I2C_ADRS 0x60

uint8_t Si5351_OEB = 0xFF;  // Clock Output Enable Control

//*******************************************************
//	Si5351A レジスタブロック書き込み
//	adrs は先頭アドレスをセットする
//  len はデータの数
//*******************************************************
static void write_reg_block(uint8_t adrs, uint8_t *buff, int len) {

  Wire.beginTransmission(SI5351_I2C_ADRS);
  Wire.write(adrs);
  for (int i = 0; i < len; i++) {
    Wire.write(buff[i]);
  }
  Wire.endTransmission();
}


//*******************************************************
//	Si5351A レジスタ書き込み
//*******************************************************
static void write_reg(uint8_t adrs, uint8_t data) {

  Wire.beginTransmission(SI5351_I2C_ADRS);
  Wire.write(adrs);
  Wire.write(data);
  Wire.endTransmission();
}


//*******************************************************
//	Si5351A レジスタ読み込み
//*******************************************************
static uint8_t read_reg(uint8_t adrs) {
  uint8_t data = 0;

  Wire.beginTransmission(SI5351_I2C_ADRS);
  Wire.write(adrs);
  Wire.endTransmission(false);

  Wire.requestFrom(SI5351_I2C_ADRS, 1);
  if (Wire.available()) {
    data = Wire.read();
  }

  return data;
}


//*******************************************************
//	Si5351A 初期化
//*******************************************************
void si5351_init(void) {

  Si5351_OEB = 0xFF;
  write_reg(SI5351_REG_OUTPUT_ENABLE, SI5351_CLK_DISABLE);  // Disable Outputs
  for (uint8_t adrs = SI5351_REG_CLK0; adrs <= SI5351_REG_CLK7; adrs++) {
    write_reg(adrs, SI5351_CLK_PDN);  // powerdown output
  }
  write_reg(SI5351_REG_INT_MASK, read_reg(SI5351_REG_INT_MASK) | 0xF0);  // Mask Interrupt

  write_reg(SI5351_REG_PLL_SRC, SI5351_CLKIN_DIV_1 | SI5351_PLLB_SRC | SI5351_PLLA_SRC);  // PLL SRC XTAL, DIV=1
  write_reg(SI5351_REG_OEB, 0xFF);                                                        // OEB not use
  //write_reg(SI5351_REG_CRYSTAL_CAP, SI5351_XTAL_CL_10PF | SI5353_XTAL_RESERVED);	// Crystal Load 10pF
  write_reg(SI5351_REG_CRYSTAL_CAP, SI5351_XTAL_CL_8PF | SI5353_XTAL_RESERVED);  // Crystal Load 8pF

#if 0
  // 動作テスト
  char msg[32];
  sprintf(msg, "Si5351:%02Xh\n", read_reg(SI5351_REG_CRYSTAL_CAP));
  Serial.write(msg);
#endif
}


//*******************************************************
//	PLL設定	mult, num, denom は a, b, c に相当する
//*******************************************************
void si5351_setupPLL(uint8_t pll, /* SI5351_PLL_A or SI5351_PLL_B */
                     uint8_t mult,
                     uint32_t num,
                     uint32_t denom) {
  /* Get the appropriate starting point for the PLL registers */
  const uint8_t pllreg_base[] = {
    SI5351_REG_26_PLL_A,
    SI5351_REG_34_PLL_B
  };

  uint32_t P1;
  uint32_t P2;
  uint32_t P3;
  uint8_t buff[9];

  /* Set the main PLL config registers */
  if (num == 0) {
    /* Integer mode */
    P1 = 128 * mult - 512;
    P2 = 0;
    P3 = 1;
  } else {
    /* Fractional mode */
    P1 = 128 * mult + ((128 * num) / denom) - 512;
    P2 = 128 * num - denom * ((128 * num) / denom);
    P3 = denom;
  }

  buff[0] = (uint8_t)((P3 & 0x0000FF00) >> 8);
  buff[1] = (uint8_t)(P3 & 0x000000FF);
  buff[2] = (uint8_t)((P1 & 0x00030000) >> 16);  // R Divice by 1
  buff[3] = (uint8_t)((P1 & 0x0000FF00) >> 8);
  buff[4] = (uint8_t)(P1 & 0x000000FF);
  buff[5] = (uint8_t)(((P3 & 0x000F0000) >> 12) | ((P2 & 0x000F0000) >> 16));
  buff[6] = (uint8_t)((P2 & 0x0000FF00) >> 8);
  buff[7] = (uint8_t)(P2 & 0x000000FF);
  write_reg_block(pllreg_base[pll], buff, 8);
}


//*******************************************************
//	マルチシンセ設定
//*******************************************************
void si5351_setupMultisynth(uint8_t output,
                            uint8_t pllSource,
                            uint32_t div,  // 4,6,8, 8+
                            uint32_t num,
                            uint32_t denom) {
  /* Get the appropriate starting point for the PLL registers */
  const uint8_t msreg_base[] = {
    SI5351_REG_MS0_PARA,
    SI5351_REG_MS1_PARA,
    SI5351_REG_MS2_PARA
  };
  const uint8_t clkctrl[] = {
    SI5351_REG_CLK0,
    SI5351_REG_CLK1,
    SI5351_REG_CLK2
  };
  uint8_t dat;

  uint32_t P1;
  uint32_t P2;
  uint32_t P3;
  uint32_t div4 = 0;
  uint8_t buff[9];

  if (div == 4) {
    div4 = SI5351_DIVBY4;
    P1 = P2 = 0;
    P3 = 1;
  } else if (num == 0) {
    /* Integer mode */
    P1 = 128 * div - 512;
    P2 = 0;
    P3 = 1;
  } else {
    /* Fractional mode */
    P1 = 128 * div + ((128 * num) / denom) - 512;
    P2 = 128 * num - denom * ((128 * num) / denom);
    P3 = denom;
  }

  buff[0] = (uint8_t)((P3 & 0x0000FF00) >> 8);
  buff[1] = (uint8_t)(P3 & 0x000000FF);
  buff[2] = (uint8_t)(((P1 & 0x00030000) >> 16) | div4);  // R Divice by 1
  buff[3] = (uint8_t)((P1 & 0x0000FF00) >> 8);
  buff[4] = (uint8_t)(P1 & 0x000000FF);
  buff[5] = (uint8_t)(((P3 & 0x000F0000) >> 12) | ((P2 & 0x000F0000) >> 16));
  buff[6] = (uint8_t)((P2 & 0x0000FF00) >> 8);
  buff[7] = (uint8_t)(P2 & 0x000000FF);
  write_reg_block(msreg_base[output], buff, 8);

  /* Configure the clk control and enable the output */
  dat = SI5351_CLK_DRIVE_STRENGTH_2MA | SI5351_CLK_INPUT_MULTISYNTH_N;
  if (pllSource == SI5351_PLL_B)
    dat |= SI5351_CLK_PLL_SELECT_B;
  if (num == 0)
    dat |= SI5351_CLK_INTEGER_MODE;
  write_reg(clkctrl[output], dat);
}


static uint32_t gcd(uint32_t x, uint32_t y) {
  uint32_t z;
  while (y != 0) {
    z = x % y;
    x = y;
    y = z;
  }
  return x;
}


//*******************************************************
//	PLL固定して分周器で周波数設定
//*******************************************************
void si5351_set_frequency_fixedpll(int channel, uint32_t pll, uint32_t pllfreq, uint32_t freq) {
  int32_t div = pllfreq / freq;  // range: 8 ~ 1800
  int32_t num = pllfreq - freq * div;
  int32_t denom = freq;
  //int32_t k = freq / (1<<20) + 1;
  int32_t k = gcd(num, denom);
  num /= k;
  denom /= k;
  while (denom >= (1L << 20)) {
    num >>= 1;
    denom >>= 1;
  }
  si5351_setupMultisynth(channel, pll, div, num, denom);
}


//*******************************************************
//	分周器固定で周波数設定
//*******************************************************
void si5351_set_frequency_fixeddiv(int channel, uint32_t pll, uint32_t freq, uint32_t div) {
  int32_t pllfreq = freq * div;
  int32_t multi = pllfreq / XTALFREQ;
  int32_t num = pllfreq - multi * XTALFREQ;
  int32_t denom = XTALFREQ;
  int32_t k = gcd(num, denom);
  num /= k;
  denom /= k;
  while (denom >= (1L << 20)) {
    num >>= 1;
    denom >>= 1;
  }
  si5351_setupPLL(pll, multi, num, denom);
  si5351_setupMultisynth(channel, pll, div, 0, 1);
}


//*******************************************************
//	出力のON/OFF
//*******************************************************
void si5351_set_out(int channel, int flag) {

  switch (channel) {
    case 0:
      if (!flag) {
        Si5351_OEB |= 1 << 0;
        write_reg(SI5351_REG_CLK0, SI5351_CLK0_PDN | SI5351_PLL_SRC | SI5351_CLK0_SRC_MS0);  // CLK0 powerdown
      } else {
        Si5351_OEB &= ~(1 << 0);
        write_reg(SI5351_REG_CLK0, SI5351_PLL_SRC | SI5351_CLK0_SRC_MS0 | SI5351_CLK0_IDRV_2mA);  // CLK0 SRC MS0 2mA
      }
      write_reg(SI5351_REG_OUTPUT_ENABLE, Si5351_OEB);
      break;
    case 1:
      if (!flag) {
        Si5351_OEB |= 1 << 1;
        write_reg(SI5351_REG_CLK1, SI5351_CLK1_PDN | SI5351_MS1_SRC_PLLA | SI5351_CLK1_SRC_MS1);  // CLK1 SRC MS1 powerdown
      } else {
        Si5351_OEB &= ~(1 << 1);
        write_reg(SI5351_REG_CLK1, SI5351_MS1_SRC_PLLA | SI5351_CLK1_SRC_MS1 | SI5351_CLK1_IDRV_2mA);  // CLK1 SRC MS1, 2mA
      }
      write_reg(SI5351_REG_OUTPUT_ENABLE, Si5351_OEB);
      break;
    case 2:
      if (!flag) {
        Si5351_OEB |= 1 << 2;
        write_reg(SI5351_REG_CLK2, SI5351_CLK2_PDN | SI5351_MS2_SRC_PLLA | SI5351_CLK2_SRC_MS2);  // CLK2 PLLA MS2, 2mA, powerdown
      } else {
        Si5351_OEB &= ~(1 << 2);
        write_reg(SI5351_REG_CLK2, SI5351_MS2_SRC_PLLA | SI5351_CLK2_SRC_MS2 | SI5351_CLK2_IDRV_2mA);  // CLK1 SRC MS1, 2mA
      }
      write_reg(SI5351_REG_OUTPUT_ENABLE, Si5351_OEB);
      break;
  }
}


//*******************************************************
//	出力がON/OFFされているか
//*******************************************************
int si5351_check_out(int channel) {

  if ((channel >= 0) && (channel <= 2)) {
    if (Si5351_OEB & (1 << channel))  // High で OFF
      return 0;
    else
      return 1;
  } else {
    return 0;
  }
}
