//*******************************************************
//
//	KTR-10
//
//	メニュー設定
//
//	2023.9.5	KAWAKAMI Yukio
//*******************************************************
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "main.h"
#include "menu.h"
#include "config.h"
#include "timer.h"
#include "lcd.h"

extern CONFIG Config;
extern volatile uint16_t MenuTimer;
extern volatile uint16_t  Timer;

typedef void (*FuncPtr)(void);

const char *MenuStr[] = {
  "1 Sig. Generate ",     // テスト用の信号を発生する
  "2 Sqlch. Level  ",     // オートスケルチが開く相対レベル設定
  "3 Sqlch. Delay  ",     // オートスケルチが閉じる時のディレイ設定
  "4 Freq. Offset  ",     // PLLの周波数誤差補正
  "5 RS9 Level     ",     // シグナル９レベルのRSSI設定
  "6 Set PTT Volt  ",     // PTT ONになる電圧を自動設定
  "7 PTT Volt High ",     // PTT ONになる電圧の上限設定
  "8 PTT Volt Low  ",     // PTT ONになる電圧の下限設定
  "9 IF Freq.      ",     // 中間周波数設定
  "10 Init Config  "      // 各種設定を初期化する
};

const FuncPtr MenuFunc[] = {
  menu_sg,
  menu_sqlch_level,
  menu_sqlch_delay,
  menu_freq_offset,
  menu_rs9_level,
  menu_set_ptt_volt,
  menu_ptt_volt_high,
  menu_ptt_volt_low,
  menu_if_freq,
  menu_init_config,
};

#define MENU_N (sizeof(MenuFunc) / sizeof(FuncPtr))

const char *NoYesMsg[] = {
  "NO ",
  "YES"
};

const char *CancelOKMsg[] = {
  "Cancel",
  "OK    "
};

const char *FreqHighLow[] = {
  "Freq > OSC",
  "OSC > Freq"
};

extern int32_t FreqStep;  // チューニングステップ

const char *TuneStepMsg = "100k  5k  1k 100";
const int32_t TuneStepValue[4] = { 100000, 5000, 1000, 100 };

const char *SGmsg[] = {
	"OFF     ",
	"50.00MHz",
	"50.20MHz",
	"50.50MHz",
	"50.62MHz",
	"50.80MHz",
	"51.00MHz",
	"54.00MHz",
};

uint16_t	SGvalue[] = {
	0,
	50000,
	50200,
	50500,
	50620,
	50800,
	51000,
	54000,
};



//*******************************************************
//	設定メニュー
//*******************************************************
void menu(void) {
  int menu;
  void (*func)(void);

  menu = 0;
  MenuTimer = MENU_TIMEOUT;
  while (MenuTimer) {
    cls();
    print("<< Menu >>");
    menu = select_menu(MenuStr, MENU_N, 0, 1, menu);
    if ((menu >= 0) && (menu < MENU_N)) {
      MenuTimer = MENU_TIMEOUT;
      func = MenuFunc[menu];
      (*func)();  // 設定関数呼び出し
    } else {
      break;
    }
  }

  // 入力状態初期化
  read_SW(SW1);
  read_SW(SW2);
  read_SW(SW3);
  read_SW(SW4);
  read_ENC();
}


//*******************************************************
//  チューニングステップ変更
//*******************************************************
void menu_tune_step(void) {
  int sel;
  int key;

  for (sel = 0; sel < 4; sel++) {
    if (FreqStep == TuneStepValue[sel]) {
      break;
    }
  }
  if (sel >= 4) sel = 2;

  print_indicator(sel, TuneStepMsg);
  locate(0, 1);
  print(TuneStepMsg);
  MenuTimer = MENU_TIMEOUT;
  while (MenuTimer) {
    sleep();
    key = inkey();
    if (key == KEY_ESC) {  // SW1
      FreqStep = TuneStepValue[0];
      return;
    } else if (key == KEY_LEFT) {  // SW2
      FreqStep = TuneStepValue[1];
      return;
    } else if (key == KEY_RIGHT) {  // SW3
      FreqStep = TuneStepValue[2];
      return;
    } else if (key == KEY_ENTER) {  // SW4
      FreqStep = TuneStepValue[3];
      return;
    } else if ((key == KEY_UP)||(key == KEY_DOWN)){
      return;  // エンコーダを回したらそのままもどる
    }
  }
}


//*******************************************************
//	受信調整用シグナルジェネレータ設定
//*******************************************************
void menu_sg(void){
	static int last_rc = 0;
	int		rc = 0;

	MenuTimer = MENU_TIMEOUT;
	cls();
	print("> Sig. Generate ");
	rc = select_menu(SGmsg, 8, 0, 1, last_rc);
	if (rc == INT_MIN)	return;

  if (rc == 0){
    set_freq(1, 0);
  } else {
    // 送信回路はOFFのまま、PLL1出力
    set_freq(1, (uint32_t)SGvalue[rc] * 1000L + (int32_t)Config.pll_offset);
  }
	last_rc = rc;
	print_ok();
}


//*******************************************************
//	周波数微調整
//*******************************************************
void menu_freq_offset(void) {
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> Freq. Offset");

  rc = select_number(0, 1, Config.pll_offset / 10, -2000, 2000, "0 Hz");
  if (rc == INT_MIN) return;
  Config.pll_offset = rc * 10;
  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	オートスケルチ レベル設定
//*******************************************************
void menu_sqlch_level(void) {
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> Sqlch. Level");

  rc = select_number(0, 1, Config.sqlch_level, 1, 50, " dB");
  if (rc == INT_MIN) return;
  Config.sqlch_level = rc;
  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	オートスケルチ OFFディレイ時間設定
//*******************************************************
void menu_sqlch_delay(void) {
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> Sqlch. Delay");

  rc = select_number(0, 1, Config.sqlch_delay, 1, 60, " Sec");
  if (rc == INT_MIN) return;
  Config.sqlch_delay = rc;
  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	RS9レベル設定
//*******************************************************
void menu_rs9_level(void) {
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> RS9 Level");

  rc = select_number(0, 1, Config.rs9_level, 10, 99, " dB");
  if (rc == INT_MIN) return;
  Config.rs9_level = rc;
  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	PTT電圧設定
//  実際に PTTスイッチを ONした時の A/D値を測定して ptt_volt_high と ptt_volt_low を決める
//*******************************************************
void menu_set_ptt_volt(void) {
  int rc;
  int key = 0;
  uint16_t ptt_volt;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("PTT ON OK?");
  while (MenuTimer) {
    sleep();

    if (!Timer){
      char  msg[20];
      Timer = 250;
      locate(0, 1);
      sprintf(msg, "%4d mV", read_ptt_volt());
      print(msg);
    }

    key = inkey();
    if (key) {
      MenuTimer = MENU_TIMEOUT;
      break;
    }
  }
  if (key != KEY_ENTER) return;

  ptt_volt = read_ptt_volt();   // ONしている間の AD_PTT 電圧読み込み、mV単位で返す

#if F_CPU >= 13000000
  Config.ptt_volt_high = ptt_volt + 150;  // +0.15V
  if (Config.ptt_volt_high > 4100)  Config.ptt_volt_high = 4100;  // 4.1V max

  if (ptt_volt < 150){
    Config.ptt_volt_low = 0;
  } else {
    Config.ptt_volt_low = ptt_volt - 150;   // -0.15V
    if (Config.ptt_volt_low > 4000){
      Config.ptt_volt_low = 4000;
    }
  }
#else
  Config.ptt_volt_high = ptt_volt + 150;  // +0.15V
  if (Config.ptt_volt_high > 3200)  Config.ptt_volt_high = 3200;  // 3.2V max

  if (ptt_volt < 150){
    Config.ptt_volt_low = 0;
  } else {
    Config.ptt_volt_low = ptt_volt - 150;   // -0.15V
    if (Config.ptt_volt_low > 3100){
      Config.ptt_volt_low = 3100;
    }
  }
#endif

  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	PTT Volt High側
//  PTT Volt High 〜 Low の間で送信する
//  Config.ptt_volt_high == 0 の場合は送信しない(受信のみ)
//*******************************************************
void menu_ptt_volt_high(void) {
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> PTT Volt High");

#if F_CPU >= 13000000
  rc = select_number(0, 1, (Config.ptt_volt_high + 50) / 100, 0, 42, "00 mV");
#else
  rc = select_number(0, 1, (Config.ptt_volt_high + 50) / 100, 0, 33, "00 mV");
#endif
  if (rc == INT_MIN) return;
  Config.ptt_volt_high = rc * 100;
  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	PTT Volt Low側
//  PTT 電圧が High 〜 Low の間で送信する
//*******************************************************
void menu_ptt_volt_low(void) {
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> PTT Volt Low");

#if F_CPU >= 13000000
  rc = select_number(0, 1, (Config.ptt_volt_low + 50) / 100, 0, 40, "00 mV");
#else
  rc = select_number(0, 1, (Config.ptt_volt_low + 50) / 100, 0, 32, "00 mV");
#endif
  if (rc == INT_MIN) return;
  Config.ptt_volt_low = rc * 100;
  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	IF周波数設定
//*******************************************************
void menu_if_freq(void) {
  int rc;
  int32_t if_freq;
  int if_mhz;
  char msg[32];

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> IF:");

  if_freq = abs(Config.if_freq / 1000);

  if (if_freq < 1L * 1000) {  // 1MHz
    sprintf(msg, "%3d kHz", if_freq);
  } else {
    sprintf(msg, "%2ld.%03ld MHz", if_freq / 1000L, if_freq % 1000L);
  }
  locate(5, 0);
  print(msg);

  locate(0, 1);
  print("Change?");
  rc = select_menu(NoYesMsg, 2, 13, 1, 0);
  if (rc == INT_MIN) return;
  if (rc == 0) return;  // NO

  cls();
  print("IF Freq.");
  rc = select_number(0, 1, if_freq / 1000, 0, 99, " MHz ");
  if (rc == INT_MIN) return;
  if_mhz = rc;

  cls();
  print("IF Freq.");
  locate(0, 1);
  if (if_mhz > 0) {
    sprintf(msg, "%2d.", if_mhz);
    print(msg);
    rc = select_number(3, 1, if_freq % 1000, 0, 999, " MHz ");
  } else {
    rc = select_number(3, 1, if_freq % 1000, 0, 999, " kHz ");
  }
  if (rc == INT_MIN) return;
  if_freq = if_mhz * 1000000L + rc * 1000L;

  cls();
  print("Freq HIGH/LOW");
  rc = select_menu(FreqHighLow, 2, 0, 1, ((Config.if_freq > 0) ? 0 : 1));
  if (rc == INT_MIN) return;

  if (rc == 1) {
    if_freq *= -1;
  }
  Config.if_freq = if_freq;

  //  debug
  //  sprintf(msg, "if_freq = %ld Hz\n", Config.if_freq);
  //  Serial.write(msg);

  save_conf(&Config);
  print_ok();
}


//*******************************************************
//	Config初期化
//*******************************************************
void menu_init_config(void){
  int rc;

  MenuTimer = MENU_TIMEOUT;
  cls();
  print("> Init Conifg");

  rc = select_menu(CancelOKMsg, 2, 0, 1, 0);
  if (rc == 1){
    init_conf(&Config);  // Config初期化
    save_conf(&Config);
    print_ok();
  }
}


//*******************************************************
//	メニュー表示・選択
//*******************************************************
int select_menu(const char *msg[], int msg_n, int x, int y, int sel) {
  int rc = INT_MIN;

  MenuTimer = MENU_TIMEOUT;
  while (MenuTimer) {
    int key = 0;
    if (msg && msg_n) {
      locate(x, y);
      print(msg[sel]);
    }
    wait_keyoff();
    while (MenuTimer) {
      sleep();
      key = inkey();
      if (key) {
        MenuTimer = MENU_TIMEOUT;
        break;
      }
    }
    if (!key) {
      break;
    } else if (key == KEY_ESC) {
      rc = INT_MIN;
      break;
    } else if (key == KEY_UP) {
      if (sel < (msg_n - 1)) sel++;
    } else if (key == KEY_DOWN) {
      if (sel > 0) --sel;
    } else if (key == KEY_ENTER) {
      rc = sel;
      break;
    }
  }
  return rc;
}


//*******************************************************
//	数値設定
//*******************************************************
int select_number(int x, int y, int value, int min, int max, const char *unit) {

  MenuTimer = MENU_TIMEOUT;
  while (MenuTimer) {
    int key = 0;
    char num[16];
    locate(x, y);
    sprintf(num, "%d%s", value, unit);
    print(num);
    if ((unit != NULL) && (unit[0])) {
      int n;
      n = 16 - x - strlen(num);
      while (n > 0) {
        print(" ");
        --n;
      }
    }
    locate(x, y);
    wait_keyoff();
    while (MenuTimer) {
      sleep();
      key = inkey();
      if (key) {
        MenuTimer = MENU_TIMEOUT;
        break;
      }
    }
    if (!key) {
      value = INT_MIN;
      break;
    } else if (key == KEY_ESC) {
      value = INT_MIN;
      break;
    } else if (key == KEY_ENTER) {
      break;
    } else if (key == KEY_UP) {
      if (value < max) {
        value++;
      }
    } else if (key == KEY_DOWN) {
      if (value > min) {
        --value;
      }
    }
  }
  return value;
}


//*******************************************************
//	OK表示
//*******************************************************
void print_ok(void) {

  locate(14, 1);
  print("OK");
  for (int i = 0; i < 1500; i++) {
    delay(1);
    if (inkey() == KEY_ESC) break;
  }
}


//*******************************************************
//	設定インジケータ表示
//*******************************************************
void print_indicator(int sel, const char *msg) {
  char buff[20];
  const char *p;
  int i, n;

  strcpy(buff, "                ");

  p = msg;
  i = 0;
  n = -1;
  while (n < sel) {
    while (i < 16) {
      if (*p != ' ') break;
      p++;
      i++;
    }
    n++;
    if (i >= 16) break;
    if (n == sel) buff[i] = 0x03;  // ┌
    p++;
    i++;
    while (i < 16) {
      if (*p == ' ') {
        if (n == sel) buff[i - 1] = 0x05;  // ┐
        break;
      }
      if (n == sel){
        if (i == 15){
          buff[i] = 0x05;  // ┐
        } else {
          buff[i] = 0x04;  // ─
        }
      }
      p++;
      i++;
    }
    if (i >= 16) break;
  }

  locate(0, 0);
  print(buff);
}


//*******************************************************
//	キー入力
//*******************************************************
int inkey(void) {

  if (read_SW(SW1)) return KEY_ESC;
  if (read_SW(SW2)) return KEY_LEFT;
  if (read_SW(SW3)) return KEY_RIGHT;
  if (read_SW(SW4)) return KEY_ENTER;

  int enc = read_ENC();
  if (enc < 0) return KEY_DOWN;
  if (enc > 0) return KEY_UP;

  return 0;
}


//*******************************************************
//	キーOFF待ち
//*******************************************************
void wait_keyoff(void) {

  while (inkey()) {
    sleep();
  }
}
